import evaluate
import nltk
import numpy as np
from datasets import Dataset
from transformers import (
    AutoConfig,
    AutoModelForSeq2SeqLM,
    AutoModelForSequenceClassification,
    AutoTokenizer,
    DataCollatorForSeq2Seq,
    Seq2SeqTrainingArguments,
    Trainer,
    TrainingArguments,
)

# todo add support for question answering task
# todo add support for fill mask
# todo add support for text_generation (causal language model)
# todo add support for text_2_text generation


def _finetune_cls(df, args):
    df = df.rename(columns={args["target"]: "labels", args["input_column"]: "text"})
    tokenizer_from = args.get("using", {}).get("tokenizer_from", args["model_name"])
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_from)
    dataset = Dataset.from_pandas(df)

    def _tokenize_text_cls_fn(examples):
        return tokenizer(examples["text"], padding="max_length", truncation=True)

    tokenized_datasets = dataset.map(_tokenize_text_cls_fn, batched=True)
    ds = tokenized_datasets.shuffle(seed=42).train_test_split(test_size=args.get("eval_size", 0.1))
    train_ds = ds["train"]
    eval_ds = ds["test"]

    ft_args = args.get("using", {}).get("trainer_args", {})
    ft_args["output_dir"] = args["model_folder"]

    n_labels = len(args["labels_map"])
    # todo replace for prod
    assert n_labels == df["labels"].nunique(), (
        f"Label mismatch! Ensure labels match what the model was originally trained on. Found {df['labels'].nunique()} classes, expected {n_labels}."
    )  # noqa
    # TODO: ideally check that labels are a subset of the original ones, too.
    config = AutoConfig.from_pretrained(args["model_name"])
    model = AutoModelForSequenceClassification.from_pretrained(args["model_name"], config=config)
    metric = evaluate.load("accuracy")
    training_args = TrainingArguments(**ft_args)

    def _compute_metrics(eval_pred):
        logits, labels = eval_pred
        predictions = np.argmax(logits, axis=-1)
        return metric.compute(predictions=predictions, references=labels)

    # generate trainer and finetune
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_ds,
        eval_dataset=eval_ds,
        compute_metrics=_compute_metrics,
    )

    return tokenizer, trainer


# TODO: merge with summarization?
def _finetune_translate(df, args):
    config = AutoConfig.from_pretrained(args["model_name"])
    df = df.rename(columns={args["target"]: "translation", args["input_column"]: "text"})
    tokenizer_from = args.get("using", {}).get("tokenizer_from", args["model_name"])
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_from)
    dataset = Dataset.from_pandas(df)

    def _tokenize_translate_fn(examples):
        source_lang = args["lang_input"]
        target_lang = args["lang_output"]
        max_target_length = config.task_specific_params["summarization"]["max_length"]
        prefix = f"translate {source_lang} to {target_lang}: "
        inputs = [prefix + ex for ex in examples["text"]]
        targets = [ex for ex in examples["translation"]]
        model_inputs = tokenizer(inputs, max_length=config.n_positions, truncation=True)

        # Setup the tokenizer for targets
        with tokenizer.as_target_tokenizer():
            labels = tokenizer(targets, max_length=max_target_length, truncation=True)

        model_inputs["labels"] = labels["input_ids"]
        return model_inputs

    tokenized_datasets = dataset.map(_tokenize_translate_fn, batched=True)
    ds = tokenized_datasets.shuffle(seed=42).train_test_split(test_size=args.get("eval_size", 0.1))
    train_ds = ds["train"]
    eval_ds = ds["test"]
    ft_args = args.get("using", {}).get("trainer_args", {})
    ft_args["output_dir"] = args["model_folder"]
    ft_args["predict_with_generate"] = True

    model = AutoModelForSeq2SeqLM.from_pretrained(args["model_name"], config=config)
    model.resize_token_embeddings(len(tokenizer))
    training_args = Seq2SeqTrainingArguments(**ft_args)
    data_collator = DataCollatorForSeq2Seq(tokenizer, model=model)

    # generate trainer and finetune
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_ds,
        eval_dataset=eval_ds,
        data_collator=data_collator,
        # compute_metrics=_compute_metrics,
    )

    return tokenizer, trainer


def _finetune_summarization(df, args):
    df = df.rename(columns={args["target"]: "summary", args["input_column"]: "text"})
    tokenizer_from = args.get("using", {}).get("tokenizer_from", args["model_name"])
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_from)
    dataset = Dataset.from_pandas(df)
    config = AutoConfig.from_pretrained(args["model_name"])

    def _tokenize_summarize_fn(examples):
        prefix = "summarize: " if "t5" in args["model_name"] else ""
        inputs = [prefix + doc for doc in examples["text"]]
        model_inputs = tokenizer(
            inputs,
            padding="max_length",
            truncation=True,
            max_length=config.max_position_embeddings,
            pad_to_max_length=True,
        )  # noqa
        labels = tokenizer(
            text_target=examples["summary"],
            max_length=config.max_position_embeddings,
            truncation=True,
        )  # noqa
        model_inputs["labels"] = labels["input_ids"]
        return model_inputs

    tokenized_datasets = dataset.map(_tokenize_summarize_fn, batched=True)
    ds = tokenized_datasets.shuffle(seed=42).train_test_split(test_size=args.get("eval_size", 0.1))
    train_ds = ds["train"]
    eval_ds = ds["test"]

    ft_args = args.get("using", {}).get("trainer_args", {})
    ft_args["output_dir"] = args["model_folder"]
    ft_args["predict_with_generate"] = True

    model = AutoModelForSeq2SeqLM.from_pretrained(args["model_name"], config=config)
    metric = evaluate.load("rouge")
    training_args = Seq2SeqTrainingArguments(**ft_args)
    data_collator = DataCollatorForSeq2Seq(tokenizer, model=model)

    def _compute_metrics(eval_pred):
        # ref: github.com/huggingface/notebooks/blob/main/examples/summarization.ipynb
        predictions, labels = eval_pred
        decoded_preds = tokenizer.batch_decode(predictions, skip_special_tokens=True)
        decoded_labels = tokenizer.batch_decode(labels, skip_special_tokens=True)

        # Rogue expects a newline after each sentence
        decoded_preds = ["\n".join(nltk.sent_tokenize(pred.strip())) for pred in decoded_preds]
        decoded_labels = ["\n".join(nltk.sent_tokenize(label.strip())) for label in decoded_labels]

        result = metric.compute(
            predictions=decoded_preds,
            references=decoded_labels,
            use_stemmer=True,
            use_aggregator=True,
        )
        result = {key: value * 100 for key, value in result.items()}
        prediction_lens = [np.count_nonzero(pred != tokenizer.pad_token_id) for pred in predictions]
        result["gen_len"] = np.mean(prediction_lens)  # todo: remove?
        return {k: round(v, 4) for k, v in result.items()}

    # generate trainer and finetune
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_ds,
        eval_dataset=eval_ds,
        data_collator=data_collator,
        compute_metrics=_compute_metrics,
    )

    return tokenizer, trainer


def _finetune_fill_mask(df, args):
    raise NotImplementedError("Finetuning fill-mask models is not yet supported.")


def _finetune_text_generation(df, args):
    raise NotImplementedError("Finetuning text-generation models is not yet supported.")


def _finetune_question_answering(df, args):
    raise NotImplementedError("Finetuning question-answering models is not yet supported.")


def _finetune_text_2_text_generation(df, args):
    raise NotImplementedError("Finetuning text-2-text generation models is not yet supported.")



from typing import List

import pandas as pd

from mindsdb.integrations.handlers.ms_teams_handler.ms_graph_api_teams_client import MSGraphAPITeamsDelegatedPermissionsClient
from mindsdb.integrations.libs.api_handler import APIResource
from mindsdb.integrations.utilities.sql_utils import (
    FilterCondition,
    FilterOperator,
    SortColumn
)


class TeamsTable(APIResource):
    """
    The table abstraction for the 'teams' resource of the Microsoft Graph API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'teams' resource of the Microsoft Graph API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        client: MSGraphAPITeamsDelegatedPermissionsClient = self.handler.connect()
        teams = client.get_teams()

        teams_df = pd.json_normalize(teams, sep="_")
        teams_df = teams_df.reindex(columns=self.get_columns(), fill_value=None)

        return teams_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'teams' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'teams' resource.
        """
        return [
            "id",
            "createdDateTime",
            "displayName",
            "description",
            "internalId",
            "classification",
            "specialization",
            "visibility",
            "webUrl",
            "isArchived",
            "tenantId",
            "isMembershipLimitedToOwners",
        ]


class ChannelsTable(APIResource):
    """
    The table abstraction for the 'channels' resource of the Microsoft Graph API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'channels' resource of the Microsoft Graph API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        client: MSGraphAPITeamsDelegatedPermissionsClient = self.handler.connect()
        channels = []

        team_id, channel_ids = None, None
        for condition in conditions:
            if condition.column == "teamId":
                if condition.op == FilterOperator.EQUAL:
                    team_id = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'teamId'."
                    )

                condition.applied = True

            if condition.column == "id":
                if condition.op == FilterOperator.EQUAL:
                    channel_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    channel_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'id'."
                    )

                condition.applied = True

        channels = client.get_channels(team_id, channel_ids)

        channels_df = pd.json_normalize(channels, sep="_")
        channels_df = channels_df[self.get_columns()]

        return channels_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'chats' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'chats' resource.
        """
        return [
            "id",
            "createdDateTime",
            "displayName",
            "description",
            "isFavoriteByDefault",
            "email",
            "tenantId",
            "webUrl",
            "membershipType",
            "teamId",
        ]


class ChannelMessagesTable(APIResource):
    """
    The table abstraction for the 'channel messages' resource of the Microsoft Graph API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'channel messages' resource of the Microsoft Graph API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        client: MSGraphAPITeamsDelegatedPermissionsClient = self.handler.connect()
        messages = []

        group_id, channel_id, message_ids = None, None, None
        for condition in conditions:
            if condition.column == "channelIdentity_teamId":
                if condition.op == FilterOperator.EQUAL:
                    group_id = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'channelIdentity_teamId'."
                    )

                condition.applied = True

            if condition.column == "channelIdentity_channelId":
                if condition.op == FilterOperator.EQUAL:
                    channel_id = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'channelIdentity_channelId'."
                    )

                condition.applied = True

            if condition.column == "id":
                if condition.op == FilterOperator.EQUAL:
                    message_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    message_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'id'."
                    )

                condition.applied = True

        if not group_id or not channel_id:
            raise ValueError("The 'channelIdentity_teamId' and 'channelIdentity_channelId' columns are required.")

        messages = client.get_channel_messages(group_id, channel_id, message_ids)

        messages_df = pd.json_normalize(messages, sep="_")
        messages_df = messages_df[self.get_columns()]

        return messages_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'chat messages' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'chat messages' resource.
        """
        return [
            "id",
            "replyToId",
            "etag",
            "messageType",
            "createdDateTime",
            "lastModifiedDateTime",
            "lastEditedDateTime",
            "deletedDateTime",
            "subject",
            "summary",
            "chatId",
            "importance",
            "locale",
            "webUrl",
            "policyViolation",
            "from_application",
            "from_device",
            "from_user_id",
            "from_user_displayName",
            "from_user_userIdentityType",
            "body_contentType",
            "body_content",
            "channelIdentity_teamId",
            "channelIdentity_channelId",
        ]


class ChatsTable(APIResource):
    """
    The table abstraction for the 'chats' resource of the Microsoft Graph API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'chats' resource of the Microsoft Graph API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        client: MSGraphAPITeamsDelegatedPermissionsClient = self.handler.connect()
        chats = []

        chat_ids = None
        for condition in conditions:
            if condition.column == "id":
                if condition.op == FilterOperator.EQUAL:
                    chat_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    chat_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'id'."
                    )

                condition.applied = True

        chats = client.get_chats(chat_ids)

        chats_df = pd.json_normalize(chats, sep="_")
        chats_df = chats_df[self.get_columns()]

        return chats_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'chats' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'chats' resource.
        """
        return [
            "id",
            "topic",
            "createdDateTime",
            "lastUpdatedDateTime",
            "chatType",
            "webUrl",
            "isHiddenForAllMembers"
        ]


class ChatMessagesTable(APIResource):
    """
    The table abstraction for the 'chat messages' resource of the Microsoft Graph API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'chat messages' resource of the Microsoft Graph API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        client: MSGraphAPITeamsDelegatedPermissionsClient = self.handler.connect()
        messages = []

        chat_id, message_ids = None, None
        for condition in conditions:
            if condition.column == "chatId":
                if condition.op == FilterOperator.EQUAL:
                    chat_id = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'chatId'."
                    )

                condition.applied = True

            if condition.column == "id":
                if condition.op == FilterOperator.EQUAL:
                    message_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    message_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'id'."
                    )

                condition.applied = True

        if not chat_id:
            raise ValueError("The 'chatId' column is required.")

        messages = client.get_chat_messages(chat_id, message_ids)

        messages_df = pd.json_normalize(messages, sep="_")
        messages_df = messages_df[self.get_columns()]

        return messages_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'chat messages' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'chat messages' resource.
        """
        return [
            "id",
            "replyToId",
            "etag",
            "messageType",
            "createdDateTime",
            "lastModifiedDateTime",
            "lastEditedDateTime",
            "deletedDateTime",
            "subject",
            "summary",
            "chatId",
            "importance",
            "locale",
            "webUrl",
            "policyViolation",
            "from_application",
            "from_device",
            "from_user_id",
            "from_user_displayName",
            "from_user_userIdentityType",
            "body_contentType",
            "body_content",
        ]



# A2A specific imports
from mindsdb.api.a2a.common.types import (
    AgentCard,
    AgentCapabilities,
    AgentSkill,
)
from mindsdb.api.a2a.common.server.server import A2AServer
from mindsdb.api.a2a.task_manager import AgentTaskManager
from mindsdb.api.a2a.agent import MindsDBAgent
from mindsdb.utilities.config import config


def get_a2a_app(
    project_name: str = "mindsdb",
):
    mindsdb_port = config.get("api", {}).get("http", {}).get("port", 47334)

    # Prepare A2A artefacts (agent card & task-manager)
    capabilities = AgentCapabilities(streaming=True)
    skill = AgentSkill(
        id="mindsdb_query",
        name="MindsDB Query",
        description="Executes natural-language queries via MindsDB agents.",
        tags=["database", "mindsdb", "query", "analytics"],
        examples=[
            "What trends exist in my sales data?",
            "Generate insights from the support tickets dataset.",
        ],
        inputModes=MindsDBAgent.SUPPORTED_CONTENT_TYPES,
        outputModes=MindsDBAgent.SUPPORTED_CONTENT_TYPES,
    )

    agent_card = AgentCard(
        name="MindsDB Agent Connector",
        description=(f"A2A connector that proxies requests to MindsDB agents in project '{project_name}'."),
        url=f"http://127.0.0.1:{mindsdb_port}",
        version="1.0.0",
        defaultInputModes=MindsDBAgent.SUPPORTED_CONTENT_TYPES,
        defaultOutputModes=MindsDBAgent.SUPPORTED_CONTENT_TYPES,
        capabilities=capabilities,
        skills=[skill],
    )

    task_manager = AgentTaskManager(
        project_name=project_name,
    )

    server = A2AServer(
        agent_card=agent_card,
        task_manager=task_manager,
    )
    return server.app



from typing import List

from atlassian import Jira
import pandas as pd

from mindsdb.integrations.libs.api_handler import APIResource
from mindsdb.integrations.utilities.sql_utils import FilterCondition, SortColumn, FilterOperator
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class JiraProjectsTable(APIResource):
    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ) -> pd.DataFrame:
        client: Jira = self.handler.connect()

        projects = []
        for condition in conditions:
            if condition.column in ('id', 'key'):
                if condition.op == FilterOperator.EQUAL:
                    projects = [client.get_project(condition.value)]
                elif condition.op == FilterOperator.IN:
                    projects = [client.get_project(project_id) for project_id in condition.value]
                else:
                    raise ValueError(f"Unsupported operator {condition.op} for column {condition.column}.")
                condition.applied = True

        if not projects:
            projects = client.get_all_projects()

        if projects:
            projects_df = pd.DataFrame(projects)
            projects_df = projects_df[self.get_columns()]
        else:
            projects_df = pd.DataFrame([], columns=self.get_columns())

        return projects_df

    def get_columns(self) -> List[str]:
        return [
            "id",
            "key",
            "name",
            "projectTypeKey",
            "simplified",
            "style",
            "isPrivate",
            "entityId",
            "uuid",
        ]


class JiraIssuesTable(APIResource):
    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ) -> pd.DataFrame:
        client: Jira = self.handler.connect()

        issues = []
        for condition in conditions:
            if condition.column in ('id', 'key'):
                if condition.op == FilterOperator.EQUAL:
                    issues = [client.get_issue(condition.value)]
                elif condition.op == FilterOperator.IN:
                    issues = [client.get_issue(issue_id) for issue_id in condition.value]
                else:
                    raise ValueError(f"Unsupported operator {condition.op} for column {condition.column}.")
                condition.applied = True

            elif condition.column in ('project_id', 'project_key', 'project_name'):
                if condition.op == FilterOperator.EQUAL:
                    issues = client.get_all_project_issues(condition.value, limit=limit)
                elif condition.op == FilterOperator.IN:
                    for project_id in condition.value:
                        issues.extend(client.get_all_project_issues(project_id, limit=limit))

                condition.applied = True

        if not issues:
            project_ids = [project['id'] for project in client.get_all_projects()]
            for project_id in project_ids:
                issues.extend(self._get_project_issues_with_limit(client, project_id, limit=limit, current_issues=issues))

        if issues:
            issues_df = self.normalize(issues)
        else:
            issues_df = pd.DataFrame([], columns=self.get_columns())

        return issues_df

    def _get_project_issues_with_limit(self, client: Jira, project_id, limit=None, current_issues=None):
        """
        Helper to get issues from a project, respecting the limit.
        """
        if current_issues is None:
            current_issues = []
        if limit:
            remaining = limit - len(current_issues)
            if remaining <= 0:
                return []
            return client.get_all_project_issues(project_id, limit=remaining)
        else:
            return client.get_all_project_issues(project_id)

    def normalize(self, issues: dict) -> pd.DataFrame:
        issues_df = pd.json_normalize(issues)
        issues_df.rename(
            columns={
                "fields.project.id": "project_id",
                "fields.project.key": "project_key",
                "fields.project.name": "project_name",
                "fields.summary": "summary",
                "fields.priority.name": "priority",
                "fields.creator.displayName": "creator",
                "fields.assignee.displayName": "assignee",
                "fields.status.name": "status",
            },
            inplace=True
        )
        issues_df = issues_df[self.get_columns()]

        return issues_df

    def get_columns(self) -> List[str]:
        return [
            "id",
            "key",
            "project_id",
            "project_key",
            "project_name",
            "summary",
            "priority",
            "creator",
            "assignee",
            "status",
        ]


class JiraGroupsTable(APIResource):
    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ) -> pd.DataFrame:
        client: Jira = self.handler.connect()

        if limit:
            groups = client.get_groups(limit=limit)['groups']
        else:
            groups = client.get_groups()['groups']

        if groups:
            groups_df = pd.DataFrame(groups)
            groups_df = groups_df[self.get_columns()]
        else:
            groups_df = pd.DataFrame([], columns=self.get_columns())

        return groups_df

    def get_columns(self) -> List[str]:
        return [
            "groupId",
            "name",
            "html",
        ]


class JiraUsersTable(APIResource):
    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ) -> pd.DataFrame:
        client: Jira = self.handler.connect()

        users = []
        for condition in conditions:
            if condition.column == 'accountId':
                if condition.op == FilterOperator.EQUAL:
                    users = [client.user(account_id=condition.value)]
                elif condition.op == FilterOperator.IN:
                    users = [client.user(account_id=accountId) for accountId in condition.value]
                else:
                    raise ValueError(f"Unsupported operator {condition.op} for column {condition.column}.")
                condition.applied = True

        if not users:
            if limit:
                users = client.users_get_all(limit=limit)
            else:
                users = client.users_get_all()

        if users:
            users_df = pd.DataFrame(users)
            users_df = users_df[self.get_columns()]
        else:
            users_df = pd.DataFrame([], columns=self.get_columns())

        return users_df

    def get_columns(self) -> List[str]:
        return [
            "accountId",
            "accountType",
            "emailAddress",
            "displayName",
            "active",
            "timeZone",
            "locale",
        ]



import copy
import os
from typing import List

from openai import OpenAI, AzureOpenAI

from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.utilities.config import config


class LLMClient:
    """
    Class for accession to LLM.
    It chooses openai client or litellm handler depending on the config
    """

    def __init__(self, llm_params: dict = None):
        params = copy.deepcopy(config.get("default_llm", {}))

        if llm_params:
            params.update(llm_params)

        self.params = params

        self.provider = params.get("provider", "openai")

        if "api_key" not in params:
            params["api_key"] = get_api_key(self.provider, params, strict=False)

        if self.provider == "azure_openai":
            azure_api_key = params.get("api_key") or os.getenv("AZURE_OPENAI_API_KEY")
            azure_api_endpoint = params.get("base_url") or os.environ.get("AZURE_OPENAI_ENDPOINT")
            azure_api_version = params.get("api_version") or os.environ.get("AZURE_OPENAI_API_VERSION")
            self._llm_client = AzureOpenAI(
                api_key=azure_api_key, azure_endpoint=azure_api_endpoint, api_version=azure_api_version, max_retries=2
            )
        elif self.provider == "openai":
            openai_api_key = params.get("api_key") or os.getenv("OPENAI_API_KEY")
            kwargs = {"api_key": openai_api_key, "max_retries": 2}
            base_url = params.get("base_url")
            if base_url:
                kwargs["base_url"] = base_url
            self.client = OpenAI(**kwargs)

        else:
            # try to use litellm
            from mindsdb.api.executor.controllers.session_controller import SessionController

            session = SessionController()
            module = session.integration_controller.get_handler_module("litellm")

            if module is None or module.Handler is None:
                raise ValueError(f'Unable to use "{self.provider}" provider. Litellm handler is not installed')

            self.client = module.Handler

    def completion(self, messages: List[dict], json_output: bool = False) -> str:
        """
        Call LLM completion and get response
        """
        params = self.params
        params["json_output"] = json_output
        if self.provider in ("azure_openai", "openai"):
            response = self.client.chat.completions.create(
                model=params["model_name"],
                messages=messages,
            )
            return response.choices[0].message.content
        else:
            kwargs = params.copy()
            model = kwargs.pop("model_name")
            kwargs.pop("provider", None)
            response = self.client.completion(self.provider, model=model, messages=messages, args=kwargs)
            return response.choices[0].message.content



import datetime as dt
from typing import Dict, Union, Any

from bson.objectid import ObjectId
from mindsdb_sql_parser.ast import Select, Update, Identifier, Star, Constant, Tuple, BinaryOperation, Latest, TypeCast
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.integrations.handlers.mongodb_handler.utils.mongodb_query import MongoQuery


class MongodbRender:
    """
    Renderer to convert SQL queries represented as ASTNodes to MongoQuery instances.
    """

    def to_mongo_query(self, node: ASTNode) -> MongoQuery:
        """
        Converts SQL query to MongoQuery instance.

        Args:
            node (ASTNode): An ASTNode representing the SQL query to be converted.

        Returns:
            MongoQuery: The converted MongoQuery instance.
        """
        if isinstance(node, Select):
            return self.select(node)
        elif isinstance(node, Update):
            return self.update(node)
        raise NotImplementedError(f"Unknown statement: {node.__class__.__name__}")

    def update(self, node: Update) -> MongoQuery:
        """
        Converts an Update statement to MongoQuery instance.

        Args:
            node (Update): An ASTNode representing the SQL Update statement.

        Returns:
            MongoQuery: The converted MongoQuery instance.
        """
        collection = node.table.parts[-1]
        mquery = MongoQuery(collection)

        filters = self.handle_where(node.where)
        row = {k: v.value for k, v in node.update_columns.items()}
        mquery.add_step({"method": "update_many", "args": [filters, {"$set": row}]})
        return mquery

    def select(self, node: Select):
        """
        Converts a Select statement to MongoQuery instance.

        Args:
            node (Select): An ASTNode representing the SQL Select statement.

        Returns:
            MongoQuery: The converted MongoQuery instance.
        """
        if not isinstance(node.from_table, Identifier):
            raise NotImplementedError(f"Not supported from {node.from_table}")

        collection = node.from_table.parts[-1]

        filters = {}

        if node.where is not None:
            filters = self.handle_where(node.where)

        group = {}
        project = {"_id": 0}  # Hide _id field when it has not been explicitly requested.
        if node.distinct:
            # Group by distinct fields.
            group = {"_id": {}}

        if node.targets is not None:
            for col in node.targets:
                if isinstance(col, Star):
                    # Show all fields.
                    project = {}
                    break
                if isinstance(col, Identifier):
                    name = col.parts[-1]
                    if col.alias is None:
                        alias = name
                    else:
                        alias = col.alias.parts[-1]

                    project[alias] = f"${name}"  # Project field.

                    # Group by distinct fields.
                    if node.distinct:
                        group["_id"][name] = f"${name}"  # Group field.
                        group[name] = {"$first": f"${name}"}  # Show field.

                elif isinstance(col, Constant):
                    val = str(col.value)  # Convert to string becuase it is interpreted as an index.
                    if col.alias is None:
                        alias = val
                    else:
                        alias = col.alias.parts[-1]
                    project[alias] = val

        if node.group_by is not None:
            # TODO
            raise NotImplementedError(f"Group {node.group_by}")

        sort = {}
        if node.order_by is not None:
            for col in node.order_by:
                name = col.field.parts[-1]
                direction = 1 if col.direction.upper() == "ASC" else -1
                sort[name] = direction

        # Compose the MongoDB query.
        mquery = MongoQuery(collection)

        method = "aggregate"
        arg = []

        # MongoDB related pipeline steps for the aggregate method.
        if node.modifiers is not None:
            for modifier in node.modifiers:
                arg.append(modifier)

        if filters:
            arg.append({"$match": filters})

        if group:
            arg.append({"$group": group})

        if project:
            arg.append({"$project": project})

        if sort:
            arg.append({"$sort": sort})

        if node.offset is not None:
            arg.append({"$skip": int(node.offset.value)})

        if node.limit is not None:
            arg.append({"$limit": int(node.limit.value)})

        mquery.add_step({"method": method, "args": [arg]})

        return mquery

    def handle_where(self, node: BinaryOperation) -> Dict:
        """
        Converts a BinaryOperation node to a dictionary of MongoDB query filters.

        Args:
            node (BinaryOperation): A BinaryOperation node representing the SQL WHERE clause.

        Returns:
            dict: The converted MongoDB query filters.
        """
        # TODO: UnaryOperation, function.
        if type(node) not in [BinaryOperation]:
            raise NotImplementedError(f"Not supported type {type(node)}")

        op = node.op.lower()
        arg1, arg2 = node.args

        if op in ("and", "or"):
            query1 = self.handle_where(arg1)
            query2 = self.handle_where(arg2)

            ops = {
                "and": "$and",
                "or": "$or",
            }
            query = {ops[op]: [query1, query2]}
            return query

        ops_map = {
            ">=": "$gte",
            ">": "$gt",
            "<": "$lt",
            "<=": "$lte",
            "<>": "$ne",
            "!=": "$ne",
            "=": "$eq",
            "==": "$eq",
            "is": "$eq",
            "is not": "$ne",
        }

        if isinstance(arg1, Identifier):
            var_name = arg1.parts[-1]
            # Simple operation.
            if isinstance(arg2, Constant):
                # Identifier and Constant.
                val = ObjectId(arg2.value) if var_name == "_id" else arg2.value
                if op in ("=", "=="):
                    pass
                elif op in ops_map:
                    op2 = ops_map[op]
                    val = {op2: val}
                else:
                    raise NotImplementedError(f"Not supported operator {op}")

                return {var_name: val}

            # IN condition.
            elif isinstance(arg2, Tuple):
                # Should be IN, NOT IN.
                ops = {"in": "$in", "not in": "$nin"}
                # Must be list of Constants.
                values = [i.value for i in arg2.items]

                if op in ops:
                    op2 = ops[op]
                    cond = {op2: values}
                else:
                    raise NotImplementedError(f"Not supported operator {op}")

                return {var_name: cond}

        # Create expression.
        val1 = self.where_element_convert(arg1)
        val2 = self.where_element_convert(arg2)

        if op in ops_map:
            op2 = ops_map[op]
        else:
            raise NotImplementedError(f"Not supported operator {op}")

        return {"$expr": {op2: [val1, val2]}}

    def where_element_convert(self, node: Union[Identifier, Latest, Constant, TypeCast]) -> Any:
        """
        Converts a WHERE element to the corresponding MongoDB query element.

        Args:
            node (Union[Identifier, Latest, Constant, TypeCast]): The WHERE element to be converted.

        Returns:
            Any: The converted MongoDB query element.

        Raises:
            NotImplementedError: If the WHERE element is not supported.
            RuntimeError: If the date format is not supported.
        """
        if isinstance(node, Identifier):
            return f"${node.parts[-1]}"
        elif isinstance(node, Latest):
            return "LATEST"
        elif isinstance(node, Constant):
            return node.value
        elif isinstance(node, TypeCast) and node.type_name.upper() in ("DATE", "DATETIME"):
            formats = ["%Y-%m-%d", "%Y-%m-%dT%H:%M:%S.%f"]
            for format in formats:
                try:
                    return dt.datetime.strptime(node.arg.value, format)
                except ValueError:
                    pass
            raise RuntimeError(f"Not supported date format. Supported: {formats}")
        else:
            raise NotImplementedError(f"Unknown where element {node}")



import datetime
from typing import Dict, List, Optional

from sqlalchemy import null, func
from sqlalchemy.orm.attributes import flag_modified

from mindsdb.interfaces.storage import db
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.utilities.config import config
from mindsdb.utilities import log


logger = log.getLogger(__name__)

default_project = config.get("default_project")


class SkillsController:
    """Handles CRUD operations at the database level for Skills"""

    def __init__(self, project_controller: ProjectController = None):
        if project_controller is None:
            project_controller = ProjectController()
        self.project_controller = project_controller

    def get_skill(
        self, skill_name: str, project_name: str = default_project, strict_case: bool = False
    ) -> Optional[db.Skills]:
        """
        Gets a skill by name. Skills are expected to have unique names.

        Parameters:
            skill_name (str): The name of the skill
            project_name (str): The name of the containing project
            strict_case (bool): If True, the skill name is case-sensitive. Defaults to False.

        Returns:
            skill (Optional[db.Skills]): The database skill object

        Raises:
            ValueError: If `project_name` does not exist
        """

        project = self.project_controller.get(name=project_name)
        query = db.Skills.query.filter(
            db.Skills.project_id == project.id,
            db.Skills.deleted_at == null(),
        )
        if strict_case:
            query = query.filter(db.Skills.name == skill_name)
        else:
            query = query.filter(func.lower(db.Skills.name) == func.lower(skill_name))

        return query.first()

    def get_skills(self, project_name: Optional[str]) -> List[dict]:
        """
        Gets all skills in a project.

        Parameters:
            project_name (Optional[str]): The name of the containing project

        Returns:
            all_skills (List[db.Skills]): List of database skill object

        Raises:
            ValueError: If `project_name` does not exist
        """

        if project_name is None:
            projects = self.project_controller.get_list()
            project_ids = list([p.id for p in projects])
        else:
            project = self.project_controller.get(name=project_name)
            project_ids = [project.id]

        query = db.session.query(db.Skills).filter(
            db.Skills.project_id.in_(project_ids), db.Skills.deleted_at == null()
        )

        return query.all()

    def add_skill(self, name: str, project_name: str, type: str, params: Dict[str, str] = {}) -> db.Skills:
        """
        Adds a skill to the database.

        Parameters:
            name (str): The name of the new skill
            project_name (str): The containing project
            type (str): The type of the skill (e.g. Knowledge Base)
            params: (Dict[str, str]): Parameters associated with the skill

        Returns:
            bot (db.Skills): The created skill

        Raises:
            ValueError: If `project_name` does not exist or skill already exists
        """
        if project_name is None:
            project_name = default_project
        project = self.project_controller.get(name=project_name)

        if not name.islower():
            raise ValueError(f"The name must be in lower case: {name}")

        skill = self.get_skill(name, project_name)

        if skill is not None:
            raise ValueError(f"Skill with name already exists: {name}")

        new_skill = db.Skills(
            name=name,
            project_id=project.id,
            type=type,
            params=params,
        )
        db.session.add(new_skill)
        db.session.commit()

        return new_skill

    def update_skill(
        self,
        skill_name: str,
        new_name: str = None,
        project_name: str = default_project,
        type: str = None,
        params: Dict[str, str] = None,
    ):
        """
        Updates an existing skill in the database.

        Parameters:
            skill_name (str): The name of the new skill, or existing skill to update
            new_name (str): Updated name of the skill
            project_name (str): The containing project
            type (str): The type of the skill (e.g. Knowledge Base)
            params: (Dict[str, str]): Parameters associated with the skill

        Returns:
            bot (db.Skills): The updated skill

        Raises:
            ValueError: If `project_name` does not exist or skill doesn't exist
        """

        existing_skill = self.get_skill(skill_name, project_name)
        if existing_skill is None:
            raise ValueError(f"Skill with name not found: {skill_name}")
        if isinstance(existing_skill.params, dict) and existing_skill.params.get("is_demo") is True:
            raise ValueError("It is forbidden to change properties of the demo object")

        if new_name is not None:
            existing_skill.name = new_name
        if type is not None:
            existing_skill.type = type
        if params is not None:
            # Merge params on update
            existing_params = {} if not existing_skill.params else existing_skill.params
            existing_params.update(params)
            # Remove None values entirely.
            params = {k: v for k, v in existing_params.items() if v is not None}
            existing_skill.params = params
            # Some versions of SQL Alchemy won't handle JSON updates correctly without this.
            # See: https://docs.sqlalchemy.org/en/20/orm/session_api.html#sqlalchemy.orm.attributes.flag_modified
            flag_modified(existing_skill, "params")

        db.session.commit()

        return existing_skill

    def delete_skill(self, skill_name: str, project_name: str = default_project, strict_case: bool = False):
        """
        Deletes a skill by name.

        Parameters:
            skill_name (str): The name of the skill to delete
            project_name (str): The name of the containing project
            strict_case (bool): If true, then skill_name is case sensitive

        Raises:
            ValueError: If `project_name` does not exist or skill doesn't exist
        """

        skill = self.get_skill(skill_name, project_name, strict_case)
        if skill is None:
            raise ValueError(f"Skill with name doesn't exist: {skill_name}")
        if isinstance(skill.params, dict) and skill.params.get("is_demo") is True:
            raise ValueError("Unable to delete demo object")
        skill.deleted_at = datetime.datetime.now()
        db.session.commit()



import time
import inspect
from dataclasses import astuple
from typing import Iterable, List

import numpy as np
import pandas as pd
from sqlalchemy.types import Integer, Float

from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb_sql_parser.ast import Insert, Identifier, CreateTable, TableColumn, DropTables

from mindsdb.api.executor.datahub.classes.response import DataHubResponse
from mindsdb.api.executor.datahub.datanodes.datanode import DataNode
from mindsdb.api.executor.datahub.classes.tables_row import TablesRow
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.api.executor.sql_query.result_set import ResultSet
from mindsdb.integrations.libs.response import HandlerResponse, INF_SCHEMA_COLUMNS_NAMES
from mindsdb.integrations.utilities.utils import get_class_name
from mindsdb.metrics import metrics
from mindsdb.utilities import log
from mindsdb.utilities.profiler import profiler
from mindsdb.utilities.exception import format_db_error_message
from mindsdb.api.executor.datahub.datanodes.system_tables import infer_mysql_type

logger = log.getLogger(__name__)


class DBHandlerException(Exception):
    pass


class IntegrationDataNode(DataNode):
    type = "integration"

    def __init__(self, integration_name, ds_type, integration_controller):
        self.integration_name = integration_name
        self.ds_type = ds_type
        self.integration_controller = integration_controller
        self.integration_handler = self.integration_controller.get_data_handler(self.integration_name)

    def get_type(self):
        return self.type

    def get_tables(self):
        response = self.integration_handler.get_tables()
        if response.type == RESPONSE_TYPE.TABLE:
            result_dict = response.data_frame.to_dict(orient="records")
            result = []
            for row in result_dict:
                result.append(TablesRow.from_dict(row))
            return result
        else:
            raise Exception(f"Can't get tables: {response.error_message}")

        result_dict = response.data_frame.to_dict(orient="records")
        return [TablesRow.from_dict(row) for row in result_dict]

    def get_table_columns_df(self, table_name: str, schema_name: str | None = None) -> pd.DataFrame:
        """Get a DataFrame containing representation of information_schema.columns for the specified table.

        Args:
            table_name (str): The name of the table to get columns from.
            schema_name (str | None): The name of the schema to get columns from.

        Returns:
            pd.DataFrame: A DataFrame containing representation of information_schema.columns for the specified table.
                          The DataFrame has list of columns as in the integrations.libs.response.INF_SCHEMA_COLUMNS_NAMES.
        """
        if "schema_name" in inspect.signature(self.integration_handler.get_columns).parameters:
            response = self.integration_handler.get_columns(table_name, schema_name)
        else:
            response = self.integration_handler.get_columns(table_name)

        if response.type == RESPONSE_TYPE.COLUMNS_TABLE:
            return response.data_frame

        if response.type != RESPONSE_TYPE.TABLE:
            logger.warning(f"Wrong response type for handler's `get_columns` call: {response.type}")
            return pd.DataFrame([], columns=astuple(INF_SCHEMA_COLUMNS_NAMES))

        # region fallback for old handlers
        df = response.data_frame
        df.columns = [name.upper() for name in df.columns]
        if "FIELD" not in df.columns or "TYPE" not in df.columns:
            logger.warning(
                f"Response from the handler's `get_columns` call does not contain required columns: f{df.columns}"
            )
            return pd.DataFrame([], columns=astuple(INF_SCHEMA_COLUMNS_NAMES))

        new_df = df[["FIELD", "TYPE"]]
        new_df.columns = ["COLUMN_NAME", "DATA_TYPE"]

        new_df[INF_SCHEMA_COLUMNS_NAMES.MYSQL_DATA_TYPE] = new_df[INF_SCHEMA_COLUMNS_NAMES.DATA_TYPE].apply(
            lambda x: infer_mysql_type(x).value
        )

        for column_name in astuple(INF_SCHEMA_COLUMNS_NAMES):
            if column_name in new_df.columns:
                continue
            new_df[column_name] = None
        # endregion

        return new_df

    def get_table_columns_names(self, table_name: str, schema_name: str | None = None) -> list[str]:
        """Get a list of column names for the specified table.

        Args:
            table_name (str): The name of the table to get columns from.
            schema_name (str | None): The name of the schema to get columns from.

        Returns:
            list[str]: A list of column names for the specified table.
        """
        df = self.get_table_columns_df(table_name, schema_name)
        return df[INF_SCHEMA_COLUMNS_NAMES.COLUMN_NAME].to_list()

    def drop_table(self, name: Identifier, if_exists=False):
        drop_ast = DropTables(tables=[name], if_exists=if_exists)
        self.query(drop_ast)

    def create_table(
        self,
        table_name: Identifier,
        result_set: ResultSet = None,
        columns: List[TableColumn] = None,
        is_replace: bool = False,
        is_create: bool = False,
        raise_if_exists: bool = True,
        **kwargs,
    ) -> DataHubResponse:
        # is_create - create table
        #   if !raise_if_exists: error will be skipped
        # is_replace - drop table if exists
        # is_create==False and is_replace==False: just insert

        table_columns_meta = {}

        if columns is None:
            columns: list[TableColumn] = result_set.get_ast_columns()
            table_columns_meta = {column.name: column.type for column in columns}

        if is_replace:
            # drop
            drop_ast = DropTables(tables=[table_name], if_exists=True)
            self.query(drop_ast)
            is_create = True

        if is_create:
            create_table_ast = CreateTable(name=table_name, columns=columns, is_replace=is_replace)
            try:
                self.query(create_table_ast)
            except Exception as e:
                if raise_if_exists:
                    raise e

        if result_set is None:
            # it is just a 'create table'
            return DataHubResponse()

        # native insert
        if hasattr(self.integration_handler, "insert"):
            df = result_set.to_df()

            result: HandlerResponse = self.integration_handler.insert(table_name.parts[-1], df)
            if result is not None:
                affected_rows = result.affected_rows
            else:
                affected_rows = None
            return DataHubResponse(affected_rows=affected_rows)

        insert_columns = [Identifier(parts=[x.alias]) for x in result_set.columns]

        # adapt table types
        for col_idx, col in enumerate(result_set.columns):
            column_type = table_columns_meta[col.alias]

            if column_type == Integer:
                type_name = "int"
            elif column_type == Float:
                type_name = "float"
            else:
                continue

            try:
                result_set.set_col_type(col_idx, type_name)
            except Exception:
                pass

        values = result_set.to_lists()

        if len(values) == 0:
            # not need to insert
            return DataHubResponse()

        insert_ast = Insert(table=table_name, columns=insert_columns, values=values, is_plain=True)

        try:
            result: DataHubResponse = self.query(insert_ast)
        except Exception as e:
            msg = f"[{self.ds_type}/{self.integration_name}]: {str(e)}"
            raise DBHandlerException(msg) from e

        return DataHubResponse(affected_rows=result.affected_rows)

    def has_support_stream(self) -> bool:
        # checks if data handler has query_stream method
        return hasattr(self.integration_handler, "query_stream") and callable(self.integration_handler.query_stream)

    @profiler.profile()
    def query_stream(self, query: ASTNode, fetch_size: int = None) -> Iterable:
        # returns generator of results from handler (split by chunks)
        return self.integration_handler.query_stream(query, fetch_size=fetch_size)

    @profiler.profile()
    def query(self, query: ASTNode | None = None, native_query: str | None = None, session=None) -> DataHubResponse:
        try:
            time_before_query = time.perf_counter()
            if query is not None:
                result: HandlerResponse = self.integration_handler.query(query)
            else:
                # try to fetch native query
                result: HandlerResponse = self.integration_handler.native_query(native_query)

            # metrics
            elapsed_seconds = time.perf_counter() - time_before_query
            query_time_with_labels = metrics.INTEGRATION_HANDLER_QUERY_TIME.labels(
                get_class_name(self.integration_handler), result.type
            )
            query_time_with_labels.observe(elapsed_seconds)

            num_rows = 0
            if result.data_frame is not None:
                num_rows = len(result.data_frame.index)
            response_size_with_labels = metrics.INTEGRATION_HANDLER_RESPONSE_SIZE.labels(
                get_class_name(self.integration_handler), result.type
            )
            response_size_with_labels.observe(num_rows)
        except Exception as e:
            msg = str(e).strip()
            if msg == "":
                msg = e.__class__.__name__
            msg = f"[{self.ds_type}/{self.integration_name}]: {msg}"
            raise DBHandlerException(msg) from e

        if result.type == RESPONSE_TYPE.ERROR:
            failed_sql_query = native_query
            if query is not None:
                failed_sql_query = query.to_string()

            raise Exception(
                format_db_error_message(
                    db_name=self.integration_handler.name,
                    db_type=self.integration_handler.__class__.name,
                    db_error_msg=result.error_message,
                    failed_query=failed_sql_query,
                )
            )

        if result.type == RESPONSE_TYPE.OK:
            return DataHubResponse(affected_rows=result.affected_rows)

        df = result.data_frame
        # region clearing df from NaN values
        # recursion error appears in pandas 1.5.3 https://github.com/pandas-dev/pandas/pull/45749
        if isinstance(df, pd.Series):
            df = df.to_frame()

        try:
            # replace python's Nan, np.NaN, np.nan and pd.NA to None
            # TODO keep all NAN to the end of processing, bacause replacing also changes dtypes
            df.replace([np.NaN, pd.NA, pd.NaT], None, inplace=True)
        except Exception as e:
            logger.error(f"Issue with clearing DF from NaN values: {e}")
        # endregion

        columns_info = [{"name": k, "type": v} for k, v in df.dtypes.items()]

        return DataHubResponse(
            data_frame=df, columns=columns_info, affected_rows=result.affected_rows, mysql_types=result.mysql_types
        )



from typing import Optional

from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

import pandas as pd
import pysqream as db

from pysqream_sqlalchemy.dialect import SqreamDialect

logger = log.getLogger(__name__)


class SQreamDBHandler(DatabaseHandler):

    name = 'sqreamdb'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """ Initialize the handler
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        self.connection_data = connection_data

        self.connection = None
        self.is_connected = False

    def connect(self):
        """
        Handles the connection to a YugabyteSQL database insance.
        """
        if self.is_connected is True:
            return self.connection

        args = {
            "database": self.connection_data.get('database'),
            "host": self.connection_data.get('host'),
            "port": self.connection_data.get('port'),
            "username": self.connection_data.get('user'),
            "password": self.connection_data.get('password'),
            "clustered": self.connection_data.get('clustered', False),
            "use_ssl": self.connection_data.get('use_ssl', False),
            "service": self.connection_data.get('service', 'sqream')
        }

        connection = db.connect(**args)

        self.is_connected = True
        self.connection = connection
        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the SQreamDB  database
        :return: success status and error message if error occurs
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            with connection.cursor() as cur:
                cur.execute('select 1;')
            response.success = True
        except db.Error as e:
            logger.error(f'Error connecting to SQreamDB {self.database}, {e}!')
            response.error_message = e

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)
        Returns:
            HandlerResponse
        """
        need_to_close = self.is_connected is False
        conn = self.connect()
        with conn.cursor() as cur:
            try:
                cur.execute(query)

                if cur.rowcount > 0 and query.upper().startswith('SELECT'):
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                self.connection.commit()
            except Exception as e:
                logger.error(f'Error running query: {query} on {self.database}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )
                self.connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement
        """
        renderer = SqlalchemyRender(SqreamDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        List all tables in SQreamDB stored in 'sqream_catalog'
        """

        query = "SELECT table_name FROM sqream_catalog.tables"

        return self.query(query)

    def get_columns(self, table_name):
        query = f"""SELECT column_name, type_name
        FROM sqream_catalog.columns
        WHERE table_name = '{table_name}';
        """
        return self.query(query)



from typing import List, Union
import pandas as pd
import json
import datetime
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.interfaces.data_catalog.base_data_catalog import BaseDataCatalog
from mindsdb.interfaces.storage import db


class DataCatalogLoader(BaseDataCatalog):
    """
    This class is responsible for loading the metadata from a data source (via the handler) and storing it in the data catalog.
    """

    def load_metadata(self) -> None:
        """
        Load the metadata from the handler and store it in the database.
        """
        if not self.is_data_catalog_supported():
            return

        loaded_table_names = self._get_loaded_table_names()

        tables = self._load_table_metadata(loaded_table_names)

        if tables:
            columns = self._load_column_metadata(tables)

            self._load_column_statistics(tables, columns)

            self._load_primary_keys(tables, columns)

            self._load_foreign_keys(tables, columns)

        self.logger.info(f"Metadata loading completed for {self.database_name}.")

    def _get_loaded_table_names(self) -> List[str]:
        """
        Retrieve the names of tables that are already present in the data catalog for the current integration.
        If table_names are provided, only those tables will be checked.

        Returns:
            List[str]: Names of tables already loaded in the data catalog.
        """
        query = db.session.query(db.MetaTables).filter_by(integration_id=self.integration_id)
        if self.table_names:
            query = query.filter(db.MetaTables.name.in_(self.table_names))

        tables = query.all()
        table_names = [table.name for table in tables]

        if table_names:
            self.logger.info(f"Tables already loaded in the data catalog: {', '.join(table_names)}.")

        return table_names

    def _load_table_metadata(self, loaded_table_names: List[str] = None) -> List[Union[db.MetaTables, None]]:
        """
        Load the table metadata from the handler.
        """
        self.logger.info(f"Loading tables for {self.database_name}")
        response = self.data_handler.meta_get_tables(self.table_names)
        if response.resp_type == RESPONSE_TYPE.ERROR:
            self.logger.error(f"Failed to load tables for {self.database_name}: {response.error_message}")
            return []
        elif response.resp_type == RESPONSE_TYPE.OK:
            self.logger.error(f"No tables found for {self.database_name}.")
            return []

        df = response.data_frame
        if df.empty:
            self.logger.info(f"No tables to add for {self.database_name}.")
            return []

        df.columns = df.columns.str.lower()

        # Filter out tables that are already loaded in the data catalog
        if loaded_table_names:
            df = df[~df["table_name"].isin(loaded_table_names)]

        if df.empty:
            self.logger.info(f"No new tables to load for {self.database_name}.")
            return []

        tables = self._add_table_metadata(df)
        self.logger.info(f"Tables loaded for {self.database_name}.")
        return tables

    def _add_table_metadata(self, df: pd.DataFrame) -> List[db.MetaTables]:
        """
        Add the table metadata to the database.
        """
        tables = []
        try:
            for row in df.to_dict(orient="records"):
                # Convert the distinct_values_count to an integer if it is not NaN, otherwise set it to None.
                val = row.get("row_count")
                row_count = int(val) if pd.notna(val) else None

                record = db.MetaTables(
                    integration_id=self.integration_id,
                    name=row.get("table_name") or row.get("name"),
                    schema=row.get("table_schema"),
                    description=row.get("table_description"),
                    type=row.get("table_type"),
                    row_count=row_count,
                )
                tables.append(record)

            db.session.add_all(tables)
            db.session.commit()
        except Exception as e:
            self.logger.error(f"Failed to add tables: {e}")
            db.session.rollback()
            raise
        return tables

    def _load_column_metadata(self, tables: db.MetaTables) -> List[db.MetaColumns]:
        """
        Load the column metadata from the handler.
        """
        self.logger.info(f"Loading columns for {self.database_name}")
        response = self.data_handler.meta_get_columns([table.name for table in tables])
        if response.resp_type == RESPONSE_TYPE.ERROR:
            self.logger.error(f"Failed to load columns for {self.database_name}: {response.error_message}")
            return []
        elif response.resp_type == RESPONSE_TYPE.OK:
            self.logger.error(f"No columns found for {self.database_name}.")
            return []

        df = response.data_frame
        if df.empty:
            self.logger.info(f"No columns to load for {self.database_name}.")
            return []

        df.columns = df.columns.str.lower()
        columns = self._add_column_metadata(df, tables)
        self.logger.info(f"Columns loaded for {self.database_name}.")
        return columns

    def _add_column_metadata(self, df: pd.DataFrame, tables: db.MetaTables) -> List[db.MetaColumns]:
        """
        Add the column metadata to the database.
        """
        columns = []
        try:
            for row in df.to_dict(orient="records"):
                record = db.MetaColumns(
                    table_id=next((table.id for table in tables if table.name == row.get("table_name"))),
                    name=row.get("column_name"),
                    data_type=row.get("data_type"),
                    default_value=row.get("column_default"),
                    description=row.get("description"),
                    is_nullable=row.get("is_nullable"),
                )
                columns.append(record)

            db.session.add_all(columns)
            db.session.commit()
        except Exception as e:
            self.logger.error(f"Failed to add columns: {e}")
            db.session.rollback()
            raise
        return columns

    def _load_column_statistics(self, tables: db.MetaTables, columns: db.MetaColumns) -> None:
        """
        Load the column statistics metadata from the handler.
        """
        self.logger.info(f"Loading column statistics for {self.database_name}")
        response = self.data_handler.meta_get_column_statistics([table.name for table in tables])
        if response.resp_type == RESPONSE_TYPE.ERROR:
            self.logger.error(f"Failed to load column statistics for {self.database_name}: {response.error_message}")
            return
        elif response.resp_type == RESPONSE_TYPE.OK:
            self.logger.error(f"No column statistics found for {self.database_name}.")
            return

        df = response.data_frame
        if df.empty:
            self.logger.info(f"No column statistics to load for {self.database_name}.")
            return

        df.columns = df.columns.str.lower()
        self._add_column_statistics(df, tables, columns)
        self.logger.info(f"Column statistics loaded for {self.database_name}.")

    def _add_column_statistics(self, df: pd.DataFrame, tables: db.MetaTables, columns: db.MetaColumns) -> None:
        """
        Add the column statistics metadata to the database.
        """
        column_statistics = []
        try:
            for row in df.to_dict(orient="records"):
                table_id = next((table.id for table in tables if table.name == row.get("table_name")))
                column_id = next(
                    (
                        column.id
                        for column in columns
                        if column.name == row.get("column_name") and column.table_id == table_id
                    )
                )

                # Convert the distinct_values_count to an integer if it is not NaN, otherwise set it to None.
                val = row.get("distinct_values_count")
                distinct_values_count = int(val) if pd.notna(val) else None
                min_val = row.get("minimum_value")
                max_val = row.get("maximum_value")

                # Convert the most_common_frequencies to a list of strings.
                most_common_frequencies = [str(val) for val in row.get("most_common_frequencies") or []]

                record = db.MetaColumnStatistics(
                    column_id=column_id,
                    most_common_values=row.get("most_common_values"),
                    most_common_frequencies=most_common_frequencies,
                    null_percentage=row.get("null_percentage"),
                    distinct_values_count=distinct_values_count,
                    minimum_value=self.to_str(min_val),
                    maximum_value=self.to_str(max_val),
                )
                column_statistics.append(record)

            db.session.add_all(column_statistics)
            db.session.commit()
        except Exception as e:
            self.logger.error(f"Failed to add column statistics: {e}")
            db.session.rollback()
            raise

    def _load_primary_keys(self, tables: db.MetaTables, columns: db.MetaColumns) -> None:
        """
        Load the primary keys metadata from the handler.
        """
        self.logger.info(f"Loading primary keys for {self.database_name}")
        response = self.data_handler.meta_get_primary_keys([table.name for table in tables])
        if response.resp_type == RESPONSE_TYPE.ERROR:
            self.logger.error(f"Failed to load primary keys for {self.database_name}: {response.error_message}")
            return
        elif response.resp_type == RESPONSE_TYPE.OK:
            self.logger.error(f"No primary keys found for {self.database_name}.")
            return

        df = response.data_frame
        if df.empty:
            self.logger.info(f"No primary keys to load for {self.database_name}.")
            return

        df.columns = df.columns.str.lower()
        self._add_primary_keys(df, tables, columns)
        self.logger.info(f"Primary keys loaded for {self.database_name}.")

    def _add_primary_keys(self, df: pd.DataFrame, tables: db.MetaTables, columns: db.MetaColumns) -> None:
        """
        Add the primary keys metadata to the database.
        """
        primary_keys = []
        try:
            for row in df.to_dict(orient="records"):
                table_id = next((table.id for table in tables if table.name == row.get("table_name")))
                column_id = next(
                    (
                        column.id
                        for column in columns
                        if column.name == row.get("column_name") and column.table_id == table_id
                    )
                )

                record = db.MetaPrimaryKeys(
                    table_id=table_id,
                    column_id=column_id,
                    constraint_name=row.get("constraint_name"),
                )
                primary_keys.append(record)

            db.session.add_all(primary_keys)
            db.session.commit()
        except Exception as e:
            self.logger.error(f"Failed to add primary keys: {e}")
            db.session.rollback()
            raise

    def _load_foreign_keys(self, tables: db.MetaTables, columns: db.MetaColumns) -> None:
        """
        Load the foreign keys metadata from the handler.
        """
        self.logger.info(f"Loading foreign keys for {self.database_name}")
        response = self.data_handler.meta_get_foreign_keys([table.name for table in tables])
        if response.resp_type == RESPONSE_TYPE.ERROR:
            self.logger.error(f"Failed to foreign keys for {self.database_name}: {response.error_message}")
            return
        elif response.resp_type == RESPONSE_TYPE.OK:
            self.logger.error(f"No foreign keys found for {self.database_name}.")
            return

        df = response.data_frame
        if df.empty:
            self.logger.info(f"No foreign keys to load for {self.database_name}.")
            return

        df.columns = df.columns.str.lower()
        self._add_foreign_keys(df, tables, columns)
        self.logger.info(f"Foreign keys loaded for {self.database_name}.")

    def _add_foreign_keys(self, df: pd.DataFrame, tables: db.MetaTables, columns: db.MetaColumns) -> None:
        """
        Add the foreign keys metadata to the database.
        """
        foreign_keys = []
        try:
            for row in df.to_dict(orient="records"):
                try:
                    parent_table_id = next((table.id for table in tables if table.name == row.get("parent_table_name")))
                    parent_column_id = next(
                        (
                            column.id
                            for column in columns
                            if column.name == row.get("parent_column_name") and column.table_id == parent_table_id
                        )
                    )
                    child_table_id = next((table.id for table in tables if table.name == row.get("child_table_name")))
                    child_column_id = next(
                        (
                            column.id
                            for column in columns
                            if column.name == row.get("child_column_name") and column.table_id == child_table_id
                        )
                    )
                except StopIteration:
                    self.logger.warning(
                        f"The foreign key relationship for {row.get('parent_table_name')} -> {row.get('child_table_name')} "
                        f"could not be established. One or more tables or columns may not exist in the metadata."
                    )
                    continue

                record = db.MetaForeignKeys(
                    parent_table_id=parent_table_id,
                    parent_column_id=parent_column_id,
                    child_table_id=child_table_id,
                    child_column_id=child_column_id,
                    constraint_name=row.get("constraint_name"),
                )
                foreign_keys.append(record)

            db.session.add_all(foreign_keys)
            db.session.commit()
        except Exception as e:
            self.logger.error(f"Failed to add foreign keys: {e}")
            db.session.rollback()
            raise

    def unload_metadata(self) -> None:
        """
        Remove the metadata for the specified database from the data catalog.
        """
        if not self.is_data_catalog_supported():
            return

        meta_tables = db.session.query(db.MetaTables).filter_by(integration_id=self.integration_id).all()

        if not meta_tables:
            self.logger.info(f"No metadata found for {self.database_name}. Nothing to remove.")
            return

        for table in meta_tables:
            db.session.query(db.MetaPrimaryKeys).filter_by(table_id=table.id).delete()
            db.session.query(db.MetaForeignKeys).filter(
                (db.MetaForeignKeys.parent_table_id == table.id) | (db.MetaForeignKeys.child_table_id == table.id)
            ).delete()
            meta_columns = db.session.query(db.MetaColumns).filter_by(table_id=table.id).all()
            for col in meta_columns:
                db.session.query(db.MetaColumnStatistics).filter_by(column_id=col.id).delete()
                db.session.delete(col)

            db.session.delete(table)
        db.session.commit()
        self.logger.info(f"Metadata for {self.database_name} removed successfully.")

    def to_str(self, val) -> str:
        """
        Convert a value to a string.
        """
        if val is None:
            return None
        if isinstance(val, (datetime.datetime, datetime.date)):
            return val.isoformat()
        if isinstance(val, (list, dict, set, tuple)):
            return json.dumps(val, default=str)
        return str(val)



from sqlalchemy import exc


def schema_id(connection, schema_name=None):
    """Fetch the id for schema"""
    cur = connection.cursor()
    if schema_name is None:
        cur.execute("SELECT current_schema")
        schema_name = cur.fetchall()[0][0]

    query = f"""
                SELECT id
                FROM sys.schemas
                WHERE name = '{schema_name}'
            """

    cur.execute(query)

    try:
        schema_id = cur.fetchall()[0][0]
    except Exception:
        raise exc.InvalidRequestError(schema_name)

    return schema_id


def table_id(connection, table_name, schema_name=None):
    """Fetch the id for schema.table_name, defaulting to current schema if
    schema is None
    """

    schema_idm = schema_id(connection=connection, schema_name=schema_name)

    q = f"""
        SELECT id
        FROM sys.tables
        WHERE name = '{table_name}'
        AND schema_id = {schema_idm}
        """

    cur = connection.cursor()
    cur.execute(q)

    try:
        table_id = cur.fetchall()[0][0]
    except Exception:
        raise exc.NoSuchTableError(table_name)

    return table_id



import os
import json
import traceback
from http import HTTPStatus
from typing import Dict, Iterable, List

from flask import request, Response
from flask_restx import Resource

from mindsdb.interfaces.agents.agents_controller import AgentsController
from mindsdb.interfaces.storage import db
from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.utilities.log import getLogger
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError


logger = getLogger(__name__)


AGENT_QUICK_RESPONSE = "I understand your request. I'm working on a detailed response for you."


def create_agent(project_name, name, agent):
    if name is None:
        return http_error(HTTPStatus.BAD_REQUEST, "Missing field", 'Missing "name" field for agent')

    if not name.islower():
        return http_error(HTTPStatus.BAD_REQUEST, "Wrong name", f"The name must be in lower case: {name}")

    model_name = agent.get("model_name")
    provider = agent.get("provider")
    skills = agent.get("skills", [])

    params = agent.get("params", {})
    if agent.get("data"):
        params["data"] = agent["data"]
    if agent.get("model"):
        params["model"] = agent["model"]
    if agent.get("prompt_template"):
        params["prompt_template"] = agent["prompt_template"]

    agents_controller = AgentsController()

    try:
        existing_agent = agents_controller.get_agent(name, project_name=project_name)
    except (ValueError, EntityNotExistsError):
        # Project must exist.
        return http_error(HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist")
    if existing_agent is not None:
        return http_error(
            HTTPStatus.CONFLICT,
            "Agent already exists",
            f"Agent with name {name} already exists. Please choose a different one.",
        )

    try:
        created_agent = agents_controller.add_agent(
            name=name, project_name=project_name, model_name=model_name, skills=skills, provider=provider, params=params
        )
        return created_agent.as_dict(), HTTPStatus.CREATED
    except (ValueError, EntityExistsError):
        # Model or skill doesn't exist.
        return http_error(
            HTTPStatus.NOT_FOUND,
            "Resource not found",
            f'The model "{model_name}" or skills "{skills}" do not exist. Please ensure that the names are correct and try again.',
        )
    except NotImplementedError:
        # Free users trying to create agent.
        return http_error(
            HTTPStatus.UNAUTHORIZED,
            "Unavailable to free users",
            f'The model "{model_name}" or skills "{skills}" do not exist. Please ensure that the names are correct and try again.',
        )


@ns_conf.route("/<project_name>/agents")
class AgentsResource(Resource):
    @ns_conf.doc("list_agents")
    @api_endpoint_metrics("GET", "/agents")
    def get(self, project_name):
        """List all agents"""
        session = SessionController()
        try:
            all_agents = session.agents_controller.get_agents(project_name)
        except EntityNotExistsError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )
        return [a.as_dict() for a in all_agents]

    @ns_conf.doc("create_agent")
    @api_endpoint_metrics("POST", "/agents")
    def post(self, project_name):
        """Create a agent"""

        # Check for required parameters.
        if "agent" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Missing parameter", 'Must provide "agent" parameter in POST body'
            )

        agent = request.json["agent"]

        name = agent.get("name")
        return create_agent(project_name, name, agent)


@ns_conf.route("/<project_name>/agents/<agent_name>")
@ns_conf.param("project_name", "Name of the project")
@ns_conf.param("agent_name", "Name of the agent")
class AgentResource(Resource):
    @ns_conf.doc("get_agent")
    @api_endpoint_metrics("GET", "/agents/agent")
    def get(self, project_name, agent_name):
        """Gets an agent by name"""
        session = SessionController()
        try:
            existing_agent = session.agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                return http_error(
                    HTTPStatus.NOT_FOUND, "Agent not found", f"Agent with name {agent_name} does not exist"
                )
            return existing_agent.as_dict()
        except (ValueError, EntityNotExistsError):
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

    @ns_conf.doc("update_agent")
    @api_endpoint_metrics("PUT", "/agents/agent")
    def put(self, project_name, agent_name):
        """Updates an agent by name, creating one if it doesn't exist"""

        # Check for required parameters.
        if "agent" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Missing parameter", 'Must provide "agent" parameter in POST body'
            )
        agents_controller = AgentsController()

        try:
            existing_agent_record = agents_controller.get_agent(agent_name, project_name=project_name)
        except (ValueError, EntityNotExistsError):
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )
        if existing_agent_record is None:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                "Creation is not allowed",
                "Creation of an agent using the PUT method is not allowed.",
            )

        agent = request.json["agent"]
        name = agent.get("name", None)

        # Agent must not exist with new name.
        if name is not None and name != agent_name:
            agent_with_new_name = agents_controller.get_agent(name, project_name=project_name)
            if agent_with_new_name is not None:
                return http_error(
                    HTTPStatus.CONFLICT,
                    "Agent already exists",
                    f"Agent with name {name} already exists. Please choose a different one.",
                )

        if existing_agent_record is None:
            # Create
            return create_agent(project_name, name, agent)

        # Update
        try:
            model_name = agent.get("model_name", None)
            skills_to_add = agent.get("skills_to_add", [])
            skills_to_remove = agent.get("skills_to_remove", [])
            skills_to_rewrite = agent.get("skills", [])
            provider = agent.get("provider")
            params = agent.get("params", {})
            if agent.get("data"):
                params["data"] = agent["data"]
            if agent.get("model"):
                params["model"] = agent["model"]
            if agent.get("prompt_template"):
                params["prompt_template"] = agent["prompt_template"]

            # Check if any of the skills to be added is of type 'retrieval'
            session = SessionController()
            skills_controller = session.skills_controller
            retrieval_skill_added = False
            if len(skills_to_add) > 0:
                skills_names = [x["name"] if isinstance(x, dict) else x for x in skills_to_add]
                retrieval_skill_added = any(
                    skills_controller.get_skill(skill_name).type == "retrieval"
                    for skill_name in skills_names
                    if skills_controller.get_skill(skill_name) is not None
                )
            elif len(skills_to_rewrite) > 0:
                retrieval_skill_added = any(
                    skills_controller.get_skill(skill_meta["name"]).type == "retrieval"
                    for skill_meta in skills_to_rewrite
                    if skills_controller.get_skill(skill_meta["name"]) is not None
                )

            if retrieval_skill_added and "mode" not in params:
                params["mode"] = "retrieval"

            updated_agent = agents_controller.update_agent(
                agent_name,
                project_name=project_name,
                name=name,
                model_name=model_name,
                skills_to_add=skills_to_add,
                skills_to_remove=skills_to_remove,
                skills_to_rewrite=skills_to_rewrite,
                provider=provider,
                params=params,
            )

            return updated_agent.as_dict()
        except EntityExistsError as e:
            return http_error(HTTPStatus.NOT_FOUND, "Resource should not exists", str(e))
        except EntityNotExistsError as e:
            # Agent or skill doesn't exist.
            return http_error(HTTPStatus.NOT_FOUND, "Resource not found", str(e))
        except ValueError as e:
            return http_error(HTTPStatus.BAD_REQUEST, "Wrong arguments", str(e))

    @ns_conf.doc("delete_agent")
    @api_endpoint_metrics("DELETE", "/agents/agent")
    def delete(self, project_name, agent_name):
        """Deletes a agent by name"""
        agents_controller = AgentsController()

        try:
            existing_agent = agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                return http_error(
                    HTTPStatus.NOT_FOUND, "Agent not found", f"Agent with name {agent_name} does not exist"
                )
        except (ValueError, EntityNotExistsError):
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        agents_controller.delete_agent(agent_name, project_name=project_name)
        return "", HTTPStatus.NO_CONTENT


def _completion_event_generator(agent_name: str, messages: List[Dict], project_name: str) -> Iterable[str]:
    logger.info(f"Starting completion event generator for agent {agent_name}")

    def json_serialize(data):
        return f"data: {json.dumps(data)}\n\n"

    try:
        # Populate API key by default if not present.
        session = SessionController()
        existing_agent = session.agents_controller.get_agent(agent_name, project_name=project_name)
        if not existing_agent.params:
            existing_agent.params = {}
        existing_agent.params["openai_api_key"] = existing_agent.params.get(
            "openai_api_key", os.getenv("OPENAI_API_KEY")
        )
        # Have to commit/flush here so DB isn't locked while streaming.
        db.session.commit()

        if "mode" not in existing_agent.params and any(
            rel.skill.type == "retrieval" for rel in existing_agent.skills_relationships
        ):
            existing_agent.params["mode"] = "retrieval"

        completion_stream = session.agents_controller.get_completion(
            existing_agent, messages, project_name=project_name, tools=[], stream=True
        )

        for chunk in completion_stream:
            if isinstance(chunk, str) and chunk.startswith("data: "):
                # The chunk is already formatted correctly, yield it as is
                yield chunk
            elif isinstance(chunk, dict):
                if "error" in chunk:
                    # Handle error chunks
                    logger.error(f"Error in completion stream: {chunk['error']}")
                    yield json_serialize({"error": chunk["error"]})
                elif chunk.get("type") == "context":
                    # Handle context message
                    yield json_serialize({"type": "context", "content": chunk.get("content")})
                elif chunk.get("type") == "sql":
                    # Handle SQL query message
                    yield json_serialize({"type": "sql", "content": chunk.get("content")})
                else:
                    # Chunk should already be formatted by agent stream.
                    yield json_serialize(chunk)
            else:
                # For any other unexpected chunk types
                yield json_serialize({"output": str(chunk)})

            logger.debug(f"Streamed chunk: {str(chunk)[:100]}...")

        logger.info("Completion stream finished")

    except Exception as e:
        logger.error(f"Error in completion event generator: {e}")
        logger.error(traceback.format_exc())
        yield json_serialize({"error": "error in completion event generator"})

    finally:
        yield json_serialize({"type": "end"})


@ns_conf.route("/<project_name>/agents/<agent_name>/completions/stream")
@ns_conf.param("project_name", "Name of the project")
@ns_conf.param("agent_name", "Name of the agent")
class AgentCompletionsStream(Resource):
    @ns_conf.doc("agent_completions_stream")
    @api_endpoint_metrics("POST", "/agents/agent/completions/stream")
    def post(self, project_name, agent_name):
        # Extract messages from request (HTTP format only)
        if "messages" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                "Missing parameter",
                'Must provide "messages" parameter in POST body',
            )

        messages = request.json["messages"]

        session = SessionController()
        try:
            existing_agent = session.agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                logger.error(f"Agent {agent_name} not found in project {project_name}")
                return http_error(
                    HTTPStatus.NOT_FOUND, "Agent not found", f"Agent with name {agent_name} does not exist"
                )
        except ValueError as e:
            logger.error(f"Project {project_name} not found: {str(e)}")
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        try:
            gen = _completion_event_generator(agent_name, messages, project_name)
            logger.info(f"Starting streaming response for agent {agent_name}")
            return Response(gen, mimetype="text/event-stream")
        except Exception as e:
            logger.error(f"Error during streaming for agent {agent_name}: {str(e)}")
            logger.error(traceback.format_exc())
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR, "Streaming error", f"An error occurred during streaming: {str(e)}"
            )


@ns_conf.route("/<project_name>/agents/<agent_name>/completions")
@ns_conf.param("project_name", "Name of the project")
@ns_conf.param("agent_name", "Name of the agent")
class AgentCompletions(Resource):
    @ns_conf.doc("agent_completions")
    @api_endpoint_metrics("POST", "/agents/agent/completions")
    def post(self, project_name, agent_name):
        """Queries an agent given a list of messages"""
        # Check for required parameters.
        if "messages" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Missing parameter", 'Must provide "messages" parameter in POST body'
            )
        agents_controller = AgentsController()

        try:
            existing_agent = agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                return http_error(
                    HTTPStatus.NOT_FOUND, "Agent not found", f"Agent with name {agent_name} does not exist"
                )
        except (ValueError, EntityNotExistsError):
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        # set mode to `retrieval` if agent has a skill of type `retrieval` and mode is not set
        if "mode" not in existing_agent.params and any(
            rel.skill.type == "retrieval" for rel in existing_agent.skills_relationships
        ):
            existing_agent.params["mode"] = "retrieval"

        messages = request.json["messages"]

        completion = agents_controller.get_completion(
            existing_agent,
            messages,
            project_name=project_name,
            # Don't need to include backoffice_db related tools into this endpoint.
            # Underlying handler (e.g. Langchain) will handle default tools like mdb_read, mdb_write, etc.
            tools=[],
        )

        output_col = agents_controller.assistant_column
        model_output = completion.iloc[-1][output_col]
        trace_id = completion.iloc[-1]["trace_id"]

        response = {"message": {"content": model_output, "role": "assistant", "trace_id": trace_id}}

        if existing_agent.params.get("return_context", False):
            context = []
            if "context" in completion.columns:
                try:
                    last_context = completion.iloc[-1]["context"]
                    if last_context:
                        context = json.loads(last_context)
                except (json.JSONDecodeError, IndexError) as e:
                    logger.error(f"Error decoding context: {e}")
                    pass  # Keeping context as an empty list in case of error

            response["message"]["context"] = context

        return response



import unittest
from mindsdb.integrations.handlers.cloud_sql_handler.cloud_sql_handler import CloudSQLHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class CloudSQLMySQLHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "host": "",
            "port": 3306,
            "user": "root",
            "password": "",
            "database": "public",
            "db_engine": "mysql"
        }
        cls.handler = CloudSQLHandler('test_cloud_sql_mysql_handler', cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_native_query_select(self):
        query = "SELECT * FROM person"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_3_get_columns(self):
        columns = self.handler.get_columns('person')
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



from enum import Enum
import json
from typing import Optional, Dict

import numpy as np
import pandas as pd
import copy

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log
from .adapters import BaseMerlionForecastAdapter, DefaultForecasterAdapter, MerlionArguments, DefaultDetectorAdapter, \
    SarimaForecasterAdapter, ProphetForecasterAdapter, MSESForecasterAdapter, IsolationForestDetectorAdapter, \
    WindStatsDetectorAdapter, ProphetDetectorAdapter

logger = log.getLogger(__name__)


class DetectorModelType(Enum):
    default = DefaultDetectorAdapter
    isolation = IsolationForestDetectorAdapter
    windstats = WindStatsDetectorAdapter
    prophet = ProphetDetectorAdapter


class ForecastModelType(Enum):
    default = DefaultForecasterAdapter
    sarima = SarimaForecasterAdapter
    prophet = ProphetForecasterAdapter
    mses = MSESForecasterAdapter


class TaskType(Enum):
    detector = DetectorModelType
    forecast = ForecastModelType


def is_invalid_type(name: str, type_class: Enum) -> bool:
    if name is None:
        return True
    return name not in type_class._member_names_


def enum_to_str(type_class: Enum) -> str:
    all = []
    for element in type_class:
        all.append(element.name)
    return "|".join(all)


def to_ts_dataframe(df: pd.DataFrame, time_col=None) -> (pd.DataFrame, str):
    columns = list(df.columns.values)
    # if time column has been specified, check the specified time column
    if time_col is not None:
        if time_col not in columns:
            raise Exception("invalid column name: " + time_col)
        if df[time_col].dtype != np.datetime64:
            try:
                idx = pd.to_datetime(df[time_col])
            except Exception as e:
                raise Exception("can not convert column to datetime: " + time_col + " " + str(e))
    # if time column has not been specified, try to find one
    else:
        datetime_cols = list(df.select_dtypes(include=["datetime"]).columns.values)
        if len(datetime_cols) > 0:
            time_col = datetime_cols[0]
            idx = pd.to_datetime(df[time_col])
        else:
            raise Exception("can not find datetime column for time series")
    # build return dataframe
    rt_df = copy.deepcopy(df)
    rt_df.drop(columns=[time_col], inplace=True)
    rt_df.index = idx
    return rt_df, time_col


class MerlionHandler(BaseMLEngine):
    name = 'merlion'

    ARG_USING_TASK = "task"
    ARG_USING_MODEL_TYPE = "model_type"
    # keys only be used to persist args to args.json
    ARG_TIME_COLUMN = "time_column"
    ARG_BASE_WINDOW = "base_window"
    ARG_PREDICT_HORIZON = "predict_horizon"
    ARG_TARGET = "target"
    ARG_COLUMN_SEQUENCE = "column_sequence"

    KWARGS_DF = "df"

    DEFAULT_MODEL_TYPE = "default"
    DEFAULT_MAX_PREDICT_STEP = 100
    DEFAULT_PREDICT_BASE_WINDOW = 10

    PERSIST_MODEL_FILE_NAME = "merlion_model"
    PERSIST_ARGS_KEY_IN_JSON_STORAGE = "args"

    def create(self, target, args=None, **kwargs):
        df: pd.DataFrame = kwargs.get(self.KWARGS_DF, None)
        # prepare arguments
        using_args = args.get("using", dict())
        task = using_args.get(self.ARG_USING_TASK, TaskType.forecast.name)
        model_type = using_args.get(self.ARG_USING_MODEL_TYPE, self.DEFAULT_MODEL_TYPE)
        timeseries_settings = args.get("timeseries_settings", dict())
        time_column = timeseries_settings.get("order_by", None)
        horizon = timeseries_settings.get("horizon", self.DEFAULT_MAX_PREDICT_STEP)
        window = timeseries_settings.get("window", self.DEFAULT_PREDICT_BASE_WINDOW)
        # update args for default value maybe has been used, only time column will be set afterwards
        serialize_args = dict()
        serialize_args[self.ARG_TARGET] = target
        serialize_args[self.ARG_USING_TASK] = task
        serialize_args[self.ARG_USING_MODEL_TYPE] = model_type
        serialize_args[self.ARG_PREDICT_HORIZON] = horizon
        serialize_args[self.ARG_BASE_WINDOW] = window

        # check df
        if df is None:
            raise Exception("missing required key in args: " + self.KWARGS_DF)
        else:
            column_sequence = sorted(list(df.columns.values))
            df = df[column_sequence]
            serialize_args[self.ARG_COLUMN_SEQUENCE] = column_sequence

        # check task, model_type and get the adapter_class
        adapter_class = self.__args_to_adapter_class(task=task, model_type=model_type)
        task_enum = TaskType[task]

        # check and cast to ts dataframe
        ts_df, time_column = to_ts_dataframe(df=df, time_col=time_column)
        serialize_args[self.ARG_TIME_COLUMN] = time_column

        # train model
        model_args = {}
        if task_enum == TaskType.forecast:
            model_args[MerlionArguments.max_forecast_steps.value] = horizon
        adapter: BaseMerlionForecastAdapter = adapter_class(**model_args)
        logger.info("Training model, args: " + json.dumps(serialize_args))
        adapter.train(df=ts_df, target=target)
        logger.info("Training model completed.")

        # persist save model
        model_bytes = adapter.to_bytes()
        self.model_storage.file_set(self.PERSIST_MODEL_FILE_NAME, model_bytes)
        self.model_storage.json_set(self.PERSIST_ARGS_KEY_IN_JSON_STORAGE, serialize_args)
        logger.info("Model and args saved.")

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:
        rt_df = df.copy(deep=True)
        # read model and args from storage
        model_bytes = self.model_storage.file_get(self.PERSIST_MODEL_FILE_NAME)
        deserialize_args = self.model_storage.json_get(self.PERSIST_ARGS_KEY_IN_JSON_STORAGE)

        # resolve args
        task = deserialize_args[self.ARG_USING_TASK]
        model_type = deserialize_args[self.ARG_USING_MODEL_TYPE]
        time_column = deserialize_args[self.ARG_TIME_COLUMN]
        target = deserialize_args[self.ARG_TARGET]
        horizon = deserialize_args[self.ARG_PREDICT_HORIZON]
        feature_column_sequence = list(deserialize_args[self.ARG_COLUMN_SEQUENCE])
        task_enum = TaskType[task]

        # check df and prepare data
        if task_enum == TaskType.forecast:
            feature_column_sequence.remove(target)
        missing_required_columns = set(feature_column_sequence) - set(rt_df.columns.values)
        if len(missing_required_columns) > 0:
            raise Exception("Missing required columns: " + ",".join(missing_required_columns))
        feature_df = rt_df[feature_column_sequence]
        ts_feature_df, _ = to_ts_dataframe(df=feature_df, time_col=time_column)

        # init model adapter
        adapter_class: BaseMerlionForecastAdapter = self.__args_to_adapter_class(task=task, model_type=model_type)
        model_args = {}
        if task_enum == TaskType.forecast:
            model_args[MerlionArguments.max_forecast_steps.value] = horizon
        adapter = adapter_class(**model_args)
        adapter.initialize_model(bytes=model_bytes)

        # predict
        pred_df = adapter.predict(df=ts_feature_df, target=target)

        # build result
        pred_df = ts_feature_df[[]].join(pred_df, how="left")

        # arrange data
        pred_df.index = rt_df.index
        if task_enum == TaskType.forecast:
            pred_df = pred_df[~pred_df[target].isna()]
            rt_df.drop(columns=[target], inplace=True)
        elif task_enum == TaskType.detector:
            pred_df[f"{target}__anomaly_score"].fillna(0, inplace=True)
        rt_df = rt_df.join(pred_df, how="right")
        return rt_df

    def __args_to_adapter_class(self, task: str, model_type: str):
        # check task_type
        try:
            task_enum = TaskType[task]
        except Exception:
            raise Exception("wrong using.task: " + task + ", valid options: " + enum_to_str(TaskType))
        # check and get model class
        try:
            adapter_class = task_enum.value[model_type].value
        except Exception as e:
            raise Exception("Wrong using.model_type: " + model_type + ", valid options: "
                            + enum_to_str(task_enum.value) + ", " + str(e))
        return adapter_class
